/**
 * @presideService
 * @singleton
 */
component {
// CONSTRUCTOR
	/**
	 * @apiWrapper.inject        TwitterApiWrapper
	 * @renderingService.inject  TwitterRenderingService
	 */
	public any function init( required any apiWrapper, required any renderingService ) {
		_setApiWrapper( arguments.apiWrapper );
		_setRenderingService( arguments.renderingService );
		return this;
	}

	public boolean function getTweets( any logger ) {
		var hasLogger = StructKeyExists( arguments, "logger" );
		var canInfo   = hasLogger && arguments.logger.canInfo();
		var canWarn   = hasLogger && arguments.logger.canWarn();
		var canError  = hasLogger && arguments.logger.canError();

		var accounts       = _getAccounts();
		var dao            = $getPresideObject( "twitter_status" );
		var tweets         = [];
		var tweet          = {};
		var maxTweetId     = "";
		var sinceTweetId   = "";
		var params         = {};
		var twitterRequest = "";
		var importComplete = false;

		if ( canInfo ) { arguments.logger.info( "#accounts.recordcount# accounts(s) found..." ); }

		for( var account in accounts ) {
			if ( canInfo ) { arguments.logger.info( "Get new tweets for [#account.screen_name#]" ); }

			maxTweetId     = "";
			sinceTweetId   = _getLatestTweetIdForAccount( account.id );
			importComplete = false;

			while( !importComplete ) {
				params = {
					  screen_name          = account.screen_name
					, count                = 200
					, tweet_mode           = "extended"
					, include_ext_alt_text = true
				};
				if ( len( maxTweetId ) ) {
					params.max_id = precisionEvaluate( "#maxTweetId# - 1" );
				}
				if ( !isnull( sinceTweetId ) ) {
					params.since_id = sinceTweetId;
				}

				twitterRequest = _getApiWrapper().getUserTimeline(
					  account = account
					, params  = params
				);

				if ( twitterRequest.getSuccess() ) {
					tweets = twitterRequest.getResponse();

					if ( canInfo ) { arguments.logger.info( "Retrieved #tweets.len()# tweet(s)" ); }

					if ( !tweets.len() ) {
						break;
					}

					for( tweet in tweets ) {
						if ( isDate( account.timeline_start_date ) && parseDateTime( tweet.created_at ) < account.timeline_start_date  ) {
							if ( canInfo ) { arguments.logger.info( "Timeline start date reached. Stopping..." ); }
							importComplete = true;
							break;
						}

						try {
							maxTweetId = tweet.id_str;
							dao.insertData( {
								  id             = tweet.id_str
								, account        = account.id
								, date_posted    = parseDateTime( tweet.created_at )
								, text           = tweet.full_text       ?: ""
								, location       = tweet.place.full_name ?: ""
								, source         = tweet.source          ?: ""
								, rendered       = _getRenderingService().renderTweet( tweet )
								, is_retweet     = !isEmpty( tweet.retweeted_status ?: {} )
								, is_reply       = !isEmpty( tweet.in_reply_to_status_id ?: "" )
								, retweeted      = tweet.retweeted       ?: false
								, favorited      = tweet.favorited       ?: false
								, retweet_count  = tweet.retweet_count   ?: 0
								, favorite_count = tweet.favorite_count  ?: 0
								, raw_data       = serializeJson( tweet )
							} );
						}
						catch( e ) {
							if ( e.NativeErrorCode ?: "" == 1062 ) {
								if ( canWarn ) { arguments.logger.warn( "Attempted to insert duplicate ID [#tweet.id_str#]" ); }
							} else {
								rethrow;
							}
						}
					}
				}
			}

			if ( canInfo ) { arguments.logger.info( "Finish getting tweets for [#account.screen_name#]" ); }
		}
		if ( canInfo ) { arguments.logger.info( "All finished." ); }

		return true;
	}

	public boolean function reRenderTweets( any logger ) {
		var hasLogger = StructKeyExists( arguments, "logger" );
		var canInfo   = hasLogger && arguments.logger.canInfo();
		var canWarn   = hasLogger && arguments.logger.canWarn();
		var canError  = hasLogger && arguments.logger.canError();

		var accounts       = _getAccounts();
		var dao            = $getPresideObject( "twitter_status" );
		var tweets         = "";
		var tweet          = "";

		if ( canInfo ) { arguments.logger.info( "#accounts.recordcount# accounts(s) found..." ); }

		for( var account in accounts ) {
			if ( canInfo ) { arguments.logger.info( "Re-render tweets for [#account.screen_name#]" ); }

			tweets = dao.selectData(
				  account      = account.id
				, selectFields = [ "id", "raw_data" ]
				, orderBy      = "id desc"
			);

			for( tweet in tweets) {
				dao.updateData( id=tweet.id, data={
					rendered = _getRenderingService().renderTweet( deserializeJson( tweet.raw_data ) )
				} );
			}

			if ( canInfo ) { arguments.logger.info( "Finish re-rendering tweets for [#account.screen_name#]" ); }
		}
		if ( canInfo ) { arguments.logger.info( "All finished." ); }

		return true;
	}

	public boolean function updateCounts( numeric days=0, any logger ) {
		var hasLogger = StructKeyExists( arguments, "logger" );
		var canInfo   = hasLogger && arguments.logger.canInfo();
		var canWarn   = hasLogger && arguments.logger.canWarn();
		var canError  = hasLogger && arguments.logger.canError();

		var accounts  = _getAccounts();
		var dao       = $getPresideObject( "twitter_status" );
		var lookups   = {};
		var tweet     = {};
		var tweetIds  = [];

		if ( canInfo ) { arguments.logger.info( "#accounts.recordcount# accounts(s) found..." ); }

		for( var account in accounts ) {
			if ( canInfo ) { arguments.logger.info( "Start updating tweet stats for [#account.screen_name#]" ); }

			tweetIds = [ 0 ];

			while( true ) {
				tweetIds = _getTweetIdsToUpdate( accountId=account.id, days=arguments.days, beforeId=tweetIds.last() );

				if ( !tweetIds.len() ) {
					break;
				}

				if ( canInfo ) { arguments.logger.info( "Updating #tweetIds.len()# tweet(s)" ); }

				var twitterRequest = _getApiWrapper().getStatuses(
					  account = account
					, params  = {
						  "id"                   = tweetIds.toList()
						, "include_entities"     = false
						, "trim_user"            = true
						, "map"                  = true
						, "include_ext_alt_text" = false
						, "include_card_uri"     = false
					}
				);

				if ( twitterRequest.getSuccess() ) {
					lookups = twitterRequest.getResponse();

					for( var id in ( lookups.id ?: {} ) ) {
						if ( isNull( lookups.id[ id ] ) ) {
							dao.deleteData( id=id );
							if ( canWarn ) { arguments.logger.warn( "Deleted tweet ID [#id#]" ); }
						} else {
							tweet = lookups.id[ id ];
							dao.updateData( id=tweet.id_str, data={
								  retweeted      = tweet.retweeted      ?: false
								, favorited      = tweet.favorited      ?: false
								, retweet_count  = tweet.retweet_count  ?: 0
								, favorite_count = tweet.favorite_count ?: 0
							} );
						}
					}
				}
			}

			if ( canInfo ) { arguments.logger.info( "Finish updating tweet stats for [#account.screen_name#]" ); }
		}

		if ( canInfo ) { arguments.logger.info( "All finished." ); }

		return true;
	}


	private numeric function _getLatestTweetIdForAccount( required string accountId ) {
		var latest = $getPresideObject( "twitter_status" ).selectData(
			  filter       = { "account.id"=arguments.accountId }
			, orderBy      = "twitter_status.id desc"
			, maxRows      = 1
			, selectFields = [ "twitter_status.id" ]
		);

		return latest.recordcount ? latest.id : nullvalue();
	}

	private array function _getTweetIdsToUpdate( required string accountId, required numeric days, required string beforeId ) {
		var filter       = "account.id = :account.id";
		var filterParams = { "account.id"=arguments.accountId };

		if ( arguments.days ) {
			filter &= " and date_posted > :date_posted";
			filterParams[ "date_posted" ] = dateAdd( "d", -arguments.days, now() );
		}
		if ( arguments.beforeId > 0 ) {
			filter &= " and twitter_status.id < :twitter_status.id";
			filterParams[ "twitter_status.id" ] = arguments.beforeId;
		}

		var tweets = $getPresideObject( "twitter_status" ).selectData(
			  filter       = filter
			, filterParams = filterParams
			, orderBy      = "twitter_status.id desc"
			, maxRows      = 100
			, selectFields = [ "twitter_status.id" ]
		);

		return tweets.valueArray( "id" );
	}

	private query function _getAccounts() {
		return $getPresideObject( "twitter_account" ).selectData( filter={ active=true } );
	}


 // GETTERS AND SETTERS
	private any function _getApiWrapper() {
		return _apiWrapper;
	}
	private void function _setApiWrapper( required any apiWrapper ) {
		_apiWrapper = arguments.apiWrapper;
	}

	private any function _getRenderingService() {
		return _renderingService;
	}
	private void function _setRenderingService( required any renderingService ) {
		_renderingService = arguments.renderingService;
	}

}