/**
 * @presideService
 * @singleton
 */
component {
// CONSTRUCTOR
	/**
	 *
	 */
	public any function init() {
		return this;
	}

	public TwitterRequest function newRequest(
		  required string method
		, required string endpoint
		, required any    account
		,          struct params = {}
	) {
		var accountConfig  = _getAccountConfig( arguments.account );
		var twitterRequest = new TwitterRequest(
			  utils                     = this
			, method                    = arguments.method
			, endpoint                  = "https://api.twitter.com/1.1/" & arguments.endpoint
			, params                    = arguments.params
			, consumerApiKey            = accountConfig.consumer_api_key        ?: ""
			, consumerApiSecretKey      = accountConfig.consumer_api_secret_key ?: ""
			, consumerAccessToken       = accountConfig.access_token            ?: ""
			, consumerAccessTokenSecret = accountConfig.access_token_secret     ?: ""
		);
		twitterRequest.buildAuthHeader().submit();

		return twitterRequest;
	}

	public void function submit( required TwitterRequest twitterRequest ) {
		var response  = "";
		var params    = arguments.twitterRequest.getParams();
		var method    = arguments.twitterRequest.getMethod();
		var paramType = method == "POST" ? "formfield" : "url";

		http url=arguments.twitterRequest.getEndpoint()
			method=method
			result="response"
			timeout="30"
		{
			httpparam type="header" name="Authorization" value=arguments.twitterRequest.getAuthHeader();

			for( var key in params ) {
				httpparam type=paramType name=key value=params[ key ];
			}
		}
		arguments.twitterRequest.setRawResponse( response );

		if ( response.status_code == 200 ) {
			arguments.twitterRequest.setSuccess( true );
			arguments.twitterRequest.setResponse( deserializeJson( response.filecontent ) );
		}
	}

	public void function buildAuthHeader( required TwitterRequest twitterRequest ) {
		var authValues = {
			  "oauth_consumer_key"     = arguments.twitterRequest.getConsumerApiKey()
			, "oauth_nonce"            = _generateNonce()
			, "oauth_signature_method" = "HMAC-SHA1"
			, "oauth_timestamp"        = _getTimestamp()
			, "oauth_token"            = arguments.twitterRequest.getConsumerAccessToken()
			, "oauth_version"          = "1.0"
		};
		var baseSignature = generateBaseSignature(
			  method         = arguments.twitterRequest.getMethod()
			, endpoint       = arguments.twitterRequest.getEndpoint()
			, authValues     = authValues
			, params         = arguments.twitterRequest.getParams()
		);
		authValues[ "oauth_signature" ] = encodeSignature( arguments.twitterRequest, baseSignature );

		var authString = encodeAndConcat( values=authValues, separator=", ", qualifier='"' );

		arguments.twitterRequest.setAuthHeader( "OAuth " & authString );
	}

	public string function encodeAndConcat( required struct values, required string separator, string qualifier="" ) {
		var parts = [];

		for( var key in arguments.values.keyArray().sort( "text" ) ) {
			parts.append( '#percentEncode( key )#=#arguments.qualifier##percentEncode( arguments.values[ key ] )##arguments.qualifier#' );
		}

		return parts.toList( arguments.separator );
	}

	public string function generateBaseSignature(
		  required string method
		, required string endpoint
		, required struct authValues
		, required struct params
	) {
		var parts     = [];
		var allParams = {};
		allParams.append( arguments.authValues );
		allParams.append( arguments.params );

		parts.append( uCase( arguments.method ) );
		parts.append( percentEncode( arguments.endpoint ) );
		parts.append( percentEncode( encodeAndConcat( values=allParams, separator="&" ) ) );

		return parts.toList( "&" );
	}

	public string function encodeSignature( required TwitterRequest twitterRequest, required string baseSignature ) {
		var encoded = hmac( arguments.baseSignature, arguments.twitterRequest.getSigningKey(), "HmacSHA1" );

		return toBase64( binaryDecode( encoded, "hex" ) );
	}

	public string function percentEncode( required string input ) {
		var chars      = listToArray( arguments.input, "" );
		var validChars = [
			  "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
			, "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z"
			, "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z"
			, "-", ".", "_", "~"
		];
		var output     = [];

		for( var char in chars ) {
			if ( validChars.find( char ) ) {
				output.append( char );
			} else {
				output.append( urlEncodedFormat( char ) )
			}
		}

		return output.toList( "" );
	}



	private string function _generateNonce() {
		var minInt = 0;
		var maxInt = createObject("java","java.lang.Integer").MAX_VALUE;
		var nonce  = _getTimestamp() & randRange( minInt, maxInt );

		return hash( nonce, "SHA" );
	}

	private numeric function _getTimestamp() {
		return int( getTickcount() / 1000 );
	}

	private any function _getAccountConfig( required any account ) {
		if ( isSimpleValue( arguments.account ) ) {
			return $getPresideObject( "twitter_account" ).selectData(
				  filter       = "id = :id or screen_name=:screen_name"
				, filterParams = { id=arguments.account, screen_name=arguments.account }
				, maxRows      = 1
			);
		}

		return arguments.account;
	}

}